#include "util.h"

// texture identifiers
GLuint yTexID;

// FBO identifier
static GLuint fb;

// handle to offscreen "window", only used to properly shut down the app
static GLuint glutWindowHandle;

/**
 * Checks for OpenGL errors.
 * Extremely useful debugging function: When developing,
 * make sure to call this after almost every GL call.
 */
void checkGLErrors (const char *label) {
    GLenum errCode;
    const GLubyte *errStr;

    if ((errCode = glGetError()) != GL_NO_ERROR) {
    	errStr = gluErrorString(errCode);
    	printf("OpenGL ERROR: ");
    	printf((char*)errStr);
    	printf("(Label: ");
    	printf(label);
    	printf(")\n.");
    }
}



/**
 * Sets up a floating point texture with NEAREST filtering.
 */
void setupTexture (const GLuint texID, int	width, int height) {
	// make active and bind
	glBindTexture(GL_TEXTURE_2D,texID);
	// turn off filtering and wrap modes
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	// define texture with floating point format
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA32F_ARB,width,height,0,GL_RGBA,GL_FLOAT,0);
	// check if that worked
	if (glGetError() != GL_NO_ERROR) {
		cerr << "glTexImage2D failed" << endl;
		exit (ERROR_TEXTURE);
	}
}


/**
 * Transfers data to texture.
 * Check web page for detailed explanation on the difference between ATI and NVIDIA.
 */
void transferToTexture (float* data, int	width, int height, GLuint texID) {
    // version (a): HW-accelerated on NVIDIA
    glBindTexture(GL_TEXTURE_2D, texID);
    glTexSubImage2D(GL_TEXTURE_2D,0,0,0,width,height,GL_RGBA,GL_FLOAT,data);
    

    // version (b): HW-accelerated on ATI
//	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, texID, 0);
//	glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
//	glRasterPos2i(0,0);
//	glDrawPixels(texSize,texSize,GL_RGBA,GL_FLOAT,data);
//	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, 0, 0);

}

/*
void createTexture(GLuint	texID, int width, int height)
{
	glTexImage2D(texID, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_FLOAT, 0);
}

*/

/**
 * error checking for GLSL
 */
void printProgramInfoLog(GLuint obj) {
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;
    glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &infologLength);
    if (infologLength > 1) {
        infoLog = (char *)malloc(infologLength);
        glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);
        printf(infoLog);
        printf("\n");
        free(infoLog);
    }
}
/**
 * prints out shader info, any errors during compilation etc.
 */
void printShaderInfoLog(GLuint obj) {
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;
    glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &infologLength);
    if (infologLength > 1) {
        infoLog = (char *)malloc(infologLength);
        glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
        printf(infoLog);
        printf("\n");
        free(infoLog);
    }
}



bool	glInit(float	*data, int	width, int height)
{
	
	int argc=1;
	char **argv=new char *[1];
	argv[0]=new char[1];
	argv[0][0]='\0';

	/**
	 * GLUT init
	 */
    glutInit ( &argc, argv );
    glutWindowHandle = glutCreateWindow("GPUShader - please bare with this window for the moment.");
    glutHideWindow();

    /*
     * GLEW init
     */
    int err = glewInit();
    // sanity check
    if (GLEW_OK != err) {
    	cout << glewGetErrorString(err) << endl;
    	return(ERROR_GLEW);
    }


    /**
     * create an off-screen rendering framebuffer object
     */

    glGenFramebuffersEXT(1, &fb);
    // bind offscreen framebuffer (that is, skip the window-specific render target)

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);


    // viewport
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // for 1:1 pixel=texture mapping
    gluOrtho2D(0.0, width, 0.0, height);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glViewport(0, 0, width, height);

    /**
     * create textures
     */

    glGenTextures (1, &yTexID);
    // set up textures
    setupTexture (yTexID, width, height);
    transferToTexture(data,width, height, yTexID);
    // set texenv mode from modulate (the default) to replace
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    // check if something went completely wrong
    checkGLErrors ("createFBOandTextures()");
    
    return true;
}



void glDeInit()
{
    // clean up
    glDeleteFramebuffersEXT(1,&fb);
    glDeleteTextures(1,&yTexID);
    glutDestroyWindow (glutWindowHandle);
}


/**
 * Checks framebuffer status.
 * Copied directly out of the spec, modified to deliver a return value.
 */
bool checkFramebufferStatus() {
    GLenum status;
    status = (GLenum) glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    switch(status) {
        case GL_FRAMEBUFFER_COMPLETE_EXT:
            return true;
        case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
	    printf("Framebuffer incomplete, incomplete attachment\n");
            return false;
        case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
	    printf("Unsupported framebuffer format\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
	    printf("Framebuffer incomplete, missing attachment\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
	    printf("Framebuffer incomplete, attached images must have same dimensions\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
	    printf("Framebuffer incomplete, attached images must have same format\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
	    printf("Framebuffer incomplete, missing draw buffer\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
	    printf("Framebuffer incomplete, missing read buffer\n");
            return false;
    }
    return false;
}

/**
 * Transfers data from currently texture, and stores it in given array.
 */
void transferFromTexture(float* data) {
    // version (a): texture is attached
    // recommended on both NVIDIA and ATI
    // glReadBuffer(attachmentpoints[readTex]);
    // glReadPixels(0, 0, 1024, 1024, GL_TEXTURE_2D,GL_FLOAT,data);

    // version b: texture is not neccessarily attached
	// ALERT : ASSUMING THE PLUGIN ALWAYS RETURNS THE TEXTURE BINDED 
	// glBindTexture(GL_TEXTURE_2D,yTexID);
	glGetTexImage(GL_TEXTURE_2D,0,GL_RGBA,GL_FLOAT,data);
}


/*
int runGl(float * _data, int _width, int _height)
{
	width = _width;
	height = _height;
	data = _data;

    glInit(argc, argv);
    createTextures();
    runShader();
    transferFromTexture(data);

    glDeInit();
    return 0;
}
*/


/*on Intel(R) machines, reads the internal timestamp counter*/
/*
extern "C" inline unsigned long long rdtsc(void){
	register unsigned long long tmp;
	__asm__ __volatile__ (
			"rdtsc"
			:"=A" (tmp)
			);
	return tmp;
}

*/
